import logging
import subprocess
import sys
import time
import traceback
from pathlib import Path
from pprint import pprint

import click
import tomlkit
from mdstream import Markdown
from rich import print
from rich.traceback import Traceback

from pyproject_pip.create import create_project
from pyproject_pip.pypip import (
    find_and_sort,
    get_package_info,
    get_requirements_packages,
    modify_pyproject_toml,
    modify_requirements,
    name_and_version,
)


@click.group(invoke_without_command=True)
@click.pass_context
@click.option(
    "-v",
    "--hatch-env",
    default=None,
    help="Specify the Hatch environment to use",
)
def cli(ctx, hatch_env) -> None:
    if ctx.invoked_subcommand is None:
        show_command(hatch_env)


@cli.command("install")
@click.argument("packages", nargs=-1)
@click.option(
    "-r",
    "--requirements",
    type=click.Path(exists=True),
    help="Install packages from the given requirements file",
)
@click.option("-U", "--upgrade", is_flag=True, help="Upgrade the package(s)")
@click.option(
    "-e",
    "--editable",
    is_flag=True,
    help="Install a package in editable mode",
)
@click.option("--hatch-env", default=None, help="Specify the Hatch environment to use")
@click.option(
    "-g",
    "--dependency-group",
    default="dependencies",
    help="Specify the dependency group to use",
)
def install_command(
    packages,
    requirements,
    upgrade,
    editable,
    hatch_env,
    dependency_group,
) -> None:
    """Install packages and update requirements.txt and pyproject.toml accordingly.

    Args:
        packages (tuple): Packages to install.
        requirements (str, optional): Path to requirements file. If provided, install packages from this file.
        upgrade (bool, optional): Whether to upgrade the packages. Defaults to False.
        editable (bool, optional): Whether to install a package in editable mode. Defaults to False.
        hatch_env (str, optional): The Hatch environment to use. Defaults to "default".
        dependency_group (str, optional): The dependency group to use. Defaults to "dependencies".
    """
    try:
        if requirements:
            if not Path(requirements).exists():
                click.echo(f"Requirements file {requirements} not found. Creating it.")
                Path(requirements).touch()
            packages = get_requirements_packages(requirements)

        for package in packages:
            package_install_cmd = [sys.executable, "-m", "pip", "install"]
            if editable:
                package_install_cmd.append("-e")
            if upgrade:
                package_install_cmd.append("-U")
            package_install_cmd.append(package)
            subprocess.check_call(package_install_cmd)

            package_name, package_version = name_and_version(package, upgrade=upgrade)
            modify_pyproject_toml(
                package,
                package_version,
                action="install",
                hatch_env=hatch_env,
                dependency_group=dependency_group,
            )
            modify_requirements(package_name, package_version, action="install")

    except subprocess.CalledProcessError as e:
        click.echo(f"Error: Failed to install {package}.", err=True)
        click.echo(f"Reason: {e}", err=True)
        sys.exit(e.returncode)


@cli.command("uninstall")
@click.argument("packages", nargs=-1)
@click.option("--hatch-env", default=None, help="Specify the Hatch environment to use")
@click.option(
    "-g",
    "--dependency-group",
    default="dependencies",
    help="Specify the dependency group to use",
)
def uninstall_command(packages, hatch_env, dependency_group) -> None:
    """Uninstall packages and update requirements.txt and pyproject.toml accordingly.

    Args:
        packages (tuple): Packages to uninstall.
        hatch_env (str, optional): The Hatch environment to use. Defaults to "default".
        dependency_group (str, optional): The dependency group to use. Defaults to "dependencies".
    """
    for package in packages:
        package_name = package.split("==")[0].split("[")[0]  # Handle extras

        try:
            subprocess.check_call(
                [sys.executable, "-m", "pip", "uninstall", package_name, "-y"],
            )
            modify_requirements(package_name, action="uninstall")
            modify_pyproject_toml(
                package_name,
                action="uninstall",
                hatch_env=hatch_env,
                dependency_group=dependency_group,
            )
            click.echo(f"Successfully uninstalled {package_name}")
        except subprocess.CalledProcessError as e:
            click.echo(f"Error: Failed to uninstall {package_name}.", err=True)
            click.echo(f"Reason: {e}", err=True)
            sys.exit(e.returncode)
        except Exception as e:
            click.echo(
                f"Unexpected error occurred while trying to uninstall {package_name}:",
                err=True,
            )
            print(Traceback.from_exception(e))
            sys.exit(1)


@cli.command("show")
@click.option("--hatch-env", default=None, help="Specify the Hatch environment to use")
def show_command(hatch_env) -> None:
    """Show the dependencies from the pyproject.toml file.

    Args:
        hatch_env (str, optional): The Hatch environment to use. Defaults to "default".
    """
    try:
        with Path("pyproject.toml").open() as f:
            content = f.read()
            pyproject = tomlkit.parse(content)

        # Determine if we are using Hatch or defaulting to project dependencies
        if "tool" in pyproject and "hatch" in pyproject["tool"] and hatch_env is not None:
            dependencies = (
                pyproject.get("tool", {}).get("hatch", {}).get("envs", {}).get(hatch_env, {}).get("dependencies", [])
            )
        else:
            dependencies = pyproject.get("project", {}).get("dependencies", [])

        if dependencies:
            click.echo("Dependencies:")
            for dep in dependencies:
                click.echo(f"  {dep}")
    except FileNotFoundError:
        print(Traceback.from_exception(FileNotFoundError))
        sys.exit(1)
    finally:
        sys.exit(0)


@cli.command("find")
@click.argument("package")
@click.option("--limit", default=5, help="Limit the number of results")
@click.option("--sort", default="downloads", help="Sort key to use")
def find_command(package, limit, sort) -> None:
    """Find a package on PyPI and optionally sort the results.

    Args:
        package (str): The package to search for.
        limit (int, optional): Limit the number of results. Defaults to 5.
        sort (str, optional): Sort key to use. Defaults to "downloads".
    """
    try:
        packages = find_and_sort(package, limit, sort)
        md = Markdown(packages)
        md.stream()
    except Exception as e:
        traceback.print_exc()

@cli.command("search") # Alias for find
@click.argument("package")
@click.option("--limit", default=10, help="Limit the number of results")
@click.option("--sort", default="downloads", help="Sort key to use")
def search_command(package, limit, sort) -> None:
    """Find a package on PyPI and optionally sort the results.

    Args:
        package (str): The package to search for.
        limit (int, optional): Limit the number of results. Defaults to 5.
        sort (str, optional): Sort key to use. Defaults to "downloads".
    """  # noqa: D205
    try:
        packages = find_and_sort(package, limit, sort)
        md = Markdown(packages)
        md.stream()
        # with md as md:
        #     md.update("# Packages found:")
        #     for p in packages:
        #         time.sleep(0.4)
        #         md.update(f"## {p['name']}")
        #         md.update(f"**Version:** {p['version']}")
        #         md.update(f"**Downloads:** {p['downloads']}")
        #         md.update(f"**Summary:** {p['summary']}")
        #         md.update(f"**URLs:** {p.get('urls', '')}")
        #     md.update("---", final=True)
    except Exception as e:
        traceback.print_exc()
    


@cli.command("info")
@click.argument("package")
@click.option("--detailed", "-d", is_flag=True, help="Show verbose output")
def info_command(package, detailed) -> None:
    """Get information about a package from PyPI.

    Args:
        package (str): The package to get information about.
        detailed (bool, optional): Show detailed output. Defaults to False.
    """
    try:
        package_info = get_package_info(package, detailed)
        md = Markdown(package_info)
        md.stream()
    except Exception as e:
        traceback.print_exc()
        sys.exit(1)


@cli.command("create")
@click.argument("project_name")
@click.argument("author")
@click.option("--description", default="", help="Project description")
@click.option("--deps", default=None, help="Dependencies separated by commas")
@click.option("--python-version", default="3.10", help="Python version to use")
@click.option("--no-cli", is_flag=True, help="Do not add a CLI")
def create_command(project_name, author, description, deps, python_version="3.10", no_cli=False) -> None:
    """Create a new Python project.

    Args:
        project_name (str): The name of the project.
        author (str): The author of the project.
        description (str, optional): The description of the project. Defaults to "".
        deps (str, optional): Dependencies separated by commas. Defaults to None.
        python_version (str, optional): Python version to use. Defaults to "3.10".
        add_cli (bool, optional): Whether to add a CLI. Defaults to True.
    """
    try:
        if deps:
            deps = deps.split(",")
        create_project(project_name, author, description, deps, python_version, not no_cli)
        click.echo(f"Project {project_name} created successfully.")
    except Exception as e:
        traceback.print_exc()
        sys.exit(1)


if __name__ == "__main__":
    cli()



"""Android."""

from __future__ import annotations

import os
import re
import sys
from functools import lru_cache
from typing import cast

from .api import PlatformDirsABC


class Android(PlatformDirsABC):
    """
    Follows the guidance `from here <https://android.stackexchange.com/a/216132>`_.

    Makes use of the `appname <platformdirs.api.PlatformDirsABC.appname>`, `version
    <platformdirs.api.PlatformDirsABC.version>`, `ensure_exists <platformdirs.api.PlatformDirsABC.ensure_exists>`.

    """

    @property
    def user_data_dir(self) -> str:
        """:return: data directory tied to the user, e.g. ``/data/user/<userid>/<packagename>/files/<AppName>``"""
        return self._append_app_name_and_version(cast(str, _android_folder()), "files")

    @property
    def site_data_dir(self) -> str:
        """:return: data directory shared by users, same as `user_data_dir`"""
        return self.user_data_dir

    @property
    def user_config_dir(self) -> str:
        """
        :return: config directory tied to the user, e.g. \
        ``/data/user/<userid>/<packagename>/shared_prefs/<AppName>``
        """
        return self._append_app_name_and_version(cast(str, _android_folder()), "shared_prefs")

    @property
    def site_config_dir(self) -> str:
        """:return: config directory shared by the users, same as `user_config_dir`"""
        return self.user_config_dir

    @property
    def user_cache_dir(self) -> str:
        """:return: cache directory tied to the user, e.g.,``/data/user/<userid>/<packagename>/cache/<AppName>``"""
        return self._append_app_name_and_version(cast(str, _android_folder()), "cache")

    @property
    def site_cache_dir(self) -> str:
        """:return: cache directory shared by users, same as `user_cache_dir`"""
        return self.user_cache_dir

    @property
    def user_state_dir(self) -> str:
        """:return: state directory tied to the user, same as `user_data_dir`"""
        return self.user_data_dir

    @property
    def user_log_dir(self) -> str:
        """
        :return: log directory tied to the user, same as `user_cache_dir` if not opinionated else ``log`` in it,
          e.g. ``/data/user/<userid>/<packagename>/cache/<AppName>/log``
        """
        path = self.user_cache_dir
        if self.opinion:
            path = os.path.join(path, "log")  # noqa: PTH118
        return path

    @property
    def user_documents_dir(self) -> str:
        """:return: documents directory tied to the user e.g. ``/storage/emulated/0/Documents``"""
        return _android_documents_folder()

    @property
    def user_downloads_dir(self) -> str:
        """:return: downloads directory tied to the user e.g. ``/storage/emulated/0/Downloads``"""
        return _android_downloads_folder()

    @property
    def user_pictures_dir(self) -> str:
        """:return: pictures directory tied to the user e.g. ``/storage/emulated/0/Pictures``"""
        return _android_pictures_folder()

    @property
    def user_videos_dir(self) -> str:
        """:return: videos directory tied to the user e.g. ``/storage/emulated/0/DCIM/Camera``"""
        return _android_videos_folder()

    @property
    def user_music_dir(self) -> str:
        """:return: music directory tied to the user e.g. ``/storage/emulated/0/Music``"""
        return _android_music_folder()

    @property
    def user_desktop_dir(self) -> str:
        """:return: desktop directory tied to the user e.g. ``/storage/emulated/0/Desktop``"""
        return "/storage/emulated/0/Desktop"

    @property
    def user_runtime_dir(self) -> str:
        """
        :return: runtime directory tied to the user, same as `user_cache_dir` if not opinionated else ``tmp`` in it,
          e.g. ``/data/user/<userid>/<packagename>/cache/<AppName>/tmp``
        """
        path = self.user_cache_dir
        if self.opinion:
            path = os.path.join(path, "tmp")  # noqa: PTH118
        return path

    @property
    def site_runtime_dir(self) -> str:
        """:return: runtime directory shared by users, same as `user_runtime_dir`"""
        return self.user_runtime_dir


@lru_cache(maxsize=1)
def _android_folder() -> str | None:
    """:return: base folder for the Android OS or None if it cannot be found"""
    try:
        # First try to get a path to android app via pyjnius
        from jnius import autoclass  # noqa: PLC0415

        context = autoclass("android.content.Context")
        result: str | None = context.getFilesDir().getParentFile().getAbsolutePath()
    except Exception:  # noqa: BLE001
        # if fails find an android folder looking a path on the sys.path
        pattern = re.compile(r"/data/(data|user/\d+)/(.+)/files")
        for path in sys.path:
            if pattern.match(path):
                result = path.split("/files")[0]
                break
        else:
            result = None
    return result


@lru_cache(maxsize=1)
def _android_documents_folder() -> str:
    """:return: documents folder for the Android OS"""
    # Get directories with pyjnius
    try:
        from jnius import autoclass  # noqa: PLC0415

        context = autoclass("android.content.Context")
        environment = autoclass("android.os.Environment")
        documents_dir: str = context.getExternalFilesDir(environment.DIRECTORY_DOCUMENTS).getAbsolutePath()
    except Exception:  # noqa: BLE001
        documents_dir = "/storage/emulated/0/Documents"

    return documents_dir


@lru_cache(maxsize=1)
def _android_downloads_folder() -> str:
    """:return: downloads folder for the Android OS"""
    # Get directories with pyjnius
    try:
        from jnius import autoclass  # noqa: PLC0415

        context = autoclass("android.content.Context")
        environment = autoclass("android.os.Environment")
        downloads_dir: str = context.getExternalFilesDir(environment.DIRECTORY_DOWNLOADS).getAbsolutePath()
    except Exception:  # noqa: BLE001
        downloads_dir = "/storage/emulated/0/Downloads"

    return downloads_dir


@lru_cache(maxsize=1)
def _android_pictures_folder() -> str:
    """:return: pictures folder for the Android OS"""
    # Get directories with pyjnius
    try:
        from jnius import autoclass  # noqa: PLC0415

        context = autoclass("android.content.Context")
        environment = autoclass("android.os.Environment")
        pictures_dir: str = context.getExternalFilesDir(environment.DIRECTORY_PICTURES).getAbsolutePath()
    except Exception:  # noqa: BLE001
        pictures_dir = "/storage/emulated/0/Pictures"

    return pictures_dir


@lru_cache(maxsize=1)
def _android_videos_folder() -> str:
    """:return: videos folder for the Android OS"""
    # Get directories with pyjnius
    try:
        from jnius import autoclass  # noqa: PLC0415

        context = autoclass("android.content.Context")
        environment = autoclass("android.os.Environment")
        videos_dir: str = context.getExternalFilesDir(environment.DIRECTORY_DCIM).getAbsolutePath()
    except Exception:  # noqa: BLE001
        videos_dir = "/storage/emulated/0/DCIM/Camera"

    return videos_dir


@lru_cache(maxsize=1)
def _android_music_folder() -> str:
    """:return: music folder for the Android OS"""
    # Get directories with pyjnius
    try:
        from jnius import autoclass  # noqa: PLC0415

        context = autoclass("android.content.Context")
        environment = autoclass("android.os.Environment")
        music_dir: str = context.getExternalFilesDir(environment.DIRECTORY_MUSIC).getAbsolutePath()
    except Exception:  # noqa: BLE001
        music_dir = "/storage/emulated/0/Music"

    return music_dir


__all__ = [
    "Android",
]



import pytest
from unittest.mock import patch, call
from pyproject_pip.create import create_project


@pytest.fixture
def mock_cwd(tmp_path):
    with patch("pyproject_pip.create.getcwd", return_value=str(tmp_path)):
        yield tmp_path


def test_create_project(mock_cwd):
    project_name = "test_project"
    author = "Test Author"
    description = "Test Description"
    deps = ["pytest", "numpy"]

    with (
        patch("pyproject_pip.create.Path.mkdir") as mock_mkdir,
        patch("pyproject_pip.create.Path.write_text") as mock_write_text,
        patch("pyproject_pip.create.Path.touch") as mock_touch,
        patch(
            "pyproject_pip.create.create_pyproject_toml",
            return_value="mock_pyproject_content",
        ) as mock_create_pyproject,
    ):
        create_project(project_name, author, description, deps)

        # Check if directories were created
        assert mock_mkdir.call_count == 8  # 5 dirs + 1 for tests
        mock_mkdir.assert_has_calls([call(exist_ok=True) for _ in range(7)], any_order=True)

        # Check if files were created with correct content
        assert mock_write_text.call_count == 9  # LICENSE, README.md, pyproject.toml, __about__.py
        mock_write_text.assert_has_calls(
            [
                call(""),  # LICENSE
                call(
                    f"# {project_name}\n\n{description}\n\n## Installation\n\n```bash\npip install {project_name}\n```\n"
                ),  # README.md
                call("mock_pyproject_content"),  # pyproject.toml
                call('__version__ = "0.0.1"'),  # __about__.py
            ],
            any_order=True,
        )

        # Check if __init__.py was touched
        assert mock_touch.call_count == 7

        # Check if create_pyproject_toml was called with correct arguments
        mock_create_pyproject.assert_called_once_with(project_name, author, description, deps, python_version="3.11", add_cli=True)


def test_create_project_with_local_deps(mock_cwd):
    with (
        patch("pyproject_pip.create.Path.mkdir"),
        patch("pyproject_pip.create.Path.write_text"),
        patch("pyproject_pip.create.Path.touch"),
        patch("pyproject_pip.create.create_pyproject_toml") as mock_create_pyproject,
    ):
        create_project(
            "local_project",
            "Local Author",
            "local",
            None,
            python_version="3.11",
            add_cli=False,
        )
        mock_create_pyproject.assert_called_once_with("local_project", "Local Author", "local", None, python_version='3.11', add_cli=False)


def test_create_project_no_deps(mock_cwd):
    with (
        patch("pyproject_pip.create.Path.mkdir"),
        patch("pyproject_pip.create.Path.write_text"),
        patch("pyproject_pip.create.Path.touch"),
        patch("pyproject_pip.create.create_pyproject_toml") as mock_create_pyproject,
    ):
        create_project("no_deps_project", "No Deps Author")
        mock_create_pyproject.assert_called_once_with(
            "no_deps_project",
            "No Deps Author",
            "",
            None,
            python_version="3.11",
            add_cli=True,
        )


def test_create_project_existing_directory(mock_cwd):
    with (
        patch("pyproject_pip.create.Path.mkdir") as mock_mkdir,
        patch("pyproject_pip.create.Path.write_text"),
        patch("pyproject_pip.create.Path.touch"),
    ):
        create_project("existing_project", "Existing Author")

        # All mkdir calls should have exist_ok=True
        for call in mock_mkdir.call_args_list:
            assert call[1].get("exist_ok", False) is True



"""
Cross-Python version compatibility.
"""

from argparse import _AppendAction as AppendAction


class ExtendAction(AppendAction):  # pragma: no-cover-gt-py37
    """
    Argparse "extend" action for Python < 3.8.
    A simplified backport from the Python standard library.
    """

    def __call__(self, parser, namespace, values, option_string=None):  # noqa: ARG002
        items = getattr(namespace, self.dest, None)
        items = [] if items is None else items[:]
        items.extend(values)
        setattr(namespace, self.dest, items)



from __future__ import annotations

import threading
import traceback
from types import TracebackType
from typing import Any
from typing import Callable
from typing import Generator
from typing import TYPE_CHECKING
import warnings

import pytest


if TYPE_CHECKING:
    from typing_extensions import Self


# Copied from cpython/Lib/test/support/threading_helper.py, with modifications.
class catch_threading_exception:
    """Context manager catching threading.Thread exception using
    threading.excepthook.

    Storing exc_value using a custom hook can create a reference cycle. The
    reference cycle is broken explicitly when the context manager exits.

    Storing thread using a custom hook can resurrect it if it is set to an
    object which is being finalized. Exiting the context manager clears the
    stored object.

    Usage:
        with threading_helper.catch_threading_exception() as cm:
            # code spawning a thread which raises an exception
            ...
            # check the thread exception: use cm.args
            ...
        # cm.args attribute no longer exists at this point
        # (to break a reference cycle)
    """

    def __init__(self) -> None:
        self.args: threading.ExceptHookArgs | None = None
        self._old_hook: Callable[[threading.ExceptHookArgs], Any] | None = None

    def _hook(self, args: threading.ExceptHookArgs) -> None:
        self.args = args

    def __enter__(self) -> Self:
        self._old_hook = threading.excepthook
        threading.excepthook = self._hook
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert self._old_hook is not None
        threading.excepthook = self._old_hook
        self._old_hook = None
        del self.args


def thread_exception_runtest_hook() -> Generator[None, None, None]:
    with catch_threading_exception() as cm:
        try:
            yield
        finally:
            if cm.args:
                thread_name = (
                    "<unknown>" if cm.args.thread is None else cm.args.thread.name
                )
                msg = f"Exception in thread {thread_name}\n\n"
                msg += "".join(
                    traceback.format_exception(
                        cm.args.exc_type,
                        cm.args.exc_value,
                        cm.args.exc_traceback,
                    )
                )
                warnings.warn(pytest.PytestUnhandledThreadExceptionWarning(msg))


@pytest.hookimpl(wrapper=True, trylast=True)
def pytest_runtest_setup() -> Generator[None, None, None]:
    yield from thread_exception_runtest_hook()


@pytest.hookimpl(wrapper=True, tryfirst=True)
def pytest_runtest_call() -> Generator[None, None, None]:
    yield from thread_exception_runtest_hook()


@pytest.hookimpl(wrapper=True, tryfirst=True)
def pytest_runtest_teardown() -> Generator[None, None, None]:
    yield from thread_exception_runtest_hook()



"""Synchronizes requirements and hatch pyproject."""

import logging
import re
import subprocess
import sys
import traceback
from pathlib import Path
from pprint import pprint

import markdown2
import requests
import tomlkit
from rich.logging import RichHandler

from pyproject_pip.create import create_project

logger = logging.getLogger(__name__)
logger.addHandler(RichHandler())

def clean_text(md_text):
    """Convert Markdown to clean plain text."""
    # Convert Markdown to HTML using markdown2
    html_description = markdown2.markdown(md_text)

    # Function to remove HTML tags
    def strip_html_tags(html):
        clean = re.compile("<.*?>")
        return re.sub(clean, "", html)

    # Strip HTML tags to get plain text
    plain_text_description = strip_html_tags(html_description)

    # Function to remove reStructuredText (reST) directives and roles
    def strip_rst(text):
        # Remove reST directives and roles
        text = re.sub(
            r"\.\. .*:: .*",
            "",
            text,
        )  # remove directives like .. image:: URL
        text = re.sub(r":\w+:.*", "", text)  # remove roles like :target: URL
        return re.sub(
            r"\.\. _.*: .*",
            "",
            text,
        )  # remove hyperlink targets like .. _name: URL

    # Clean the plain text description from reST and other special characters
    clean_text = strip_rst(plain_text_description)
    clean_text = re.sub(r"&nbsp;", " ", clean_text)  # Replace HTML entities
    clean_text = re.sub(r'\n{2,}', '\n\n', clean_text)  # Replace multiple newlines with a single newline
    return clean_text.strip()


def get_latest_version(package_name):
    """Gets the latest version of the specified package from PyPI.

    Args:
        package_name (str): The name of the package to fetch the latest version for.

    Returns:
        str or None: The latest version of the package, or None if not found or on error.
    """
    try:
        response = requests.get(f"https://pypi.org/pypi/{package_name}/json")
        response.raise_for_status()  # Raises stored HTTPError, if one occurred.
        data = response.json()
        releases = data.get("releases", {})
        if releases:
            filtered_versions = [version for version in releases if all(part.isdigit() for part in version.split("."))]
            sorted_versions = sorted(
                filtered_versions,
                key=lambda x: tuple(map(int, x.split("."))),
                reverse=True,
            )
            return sorted_versions[0] if sorted_versions else None
    except Exception:
        pass
    return None


def search_package(package_name):
    """Search for a package on PyPI and return the description, details, and GitHub URL if available.

    Args:
        package_name (str): The name of the package to search for.

    Returns:
        dict: The package information.
    """
    package_info = {}
    try:
        response = requests.get(f"https://pypi.org/pypi/{package_name}/json", timeout=5)
        response.raise_for_status()  # Raises stored HTTPError, if one occurred.
        data = response.json()
        info = data.get("info", {})

        package_info["description"] = info.get("summary", "")
        package_info["details"] = info.get("description", "")
        logging.debug("Package : %s", package_info)
        # Get GitHub URL if available
        project_urls = info.get("project_urls", {})
        package_info["github_url"] = next(
            (url for _, url in project_urls.items() if "github.com" in url.lower()),
            None,
        )
    except requests.RequestException:
        pass
    except Exception:
        logging.exception("Error fetching package info: %s", traceback.format_exc())

    return package_info


def get_package_names(query_key):
    """Fetch package names from PyPI search results."""
    search_url = f"https://pypi.org/search/?q={query_key}"
    response = requests.get(search_url, timeout=20)
    response.raise_for_status()
    page_content = response.text

    # Extract package names from search results
    start_token = '<a class="package-snippet"'
    end_token = "</a>"
    name_token = '<span class="package-snippet__name">'
        
    package_names = []
    start = 0
    while True:
        start = page_content.find(start_token, start)
        if start == -1:
            break
        end = page_content.find(end_token, start)
        snippet = page_content[start:end]
        name_start = snippet.find(name_token)
        if name_start != -1:
            name_start += len(name_token)
            name_end = snippet.find("</span>", name_start)
            package_name = snippet[name_start:name_end]
            package_names.append(package_name)
        start = end
    return package_names


def get_package_info(package_name, verbose=False):
    """Retrieve detailed package information from PyPI JSON API."""
    package_url = f"https://pypi.org/pypi/{package_name}/json"
    response = requests.get(package_url, timeout=5)
    response.raise_for_status()
    package_data = response.json()
    info = package_data.get("info", {})

    downloads = package_data.get("downloads", {}).get("last_month", 0)

    package_info = {
        "name": info.get("name", ""),
        "version": info.get("version", ""),
        "summary": info.get("summary", ""),
        "downloads": downloads,
        "urls": info.get("project_urls", {}),
    }
    if verbose:
        package_info["description"] = info.get("description", "")
    project_urls = info.get("project_urls", {})
    try:
        package_info["github_url"] = next(
            (url for _, url in project_urls.items() if "github.com" in url.lower()),
            None,
        )
    except (StopIteration, TypeError, AttributeError):
        package_info["github_url"] = None

    return package_info


def find_and_sort(query_key, limit=5, sort_key="downloads") -> list:
    """Find and sort potential packages by a specified key.

    Args:
        query_key (str): The key to query for.
        sort_key (str): The key to sort by. Defaults to "downloads".

    Returns:
        list: List of packages sorted by the specified key.
    """
    try:
        package_names = get_package_names(query_key)
        packages = []
        for package_name in package_names:
            package_info = get_package_info(package_name)
            packages.append(package_info)
        return packages[:limit]
        # Sort the packages by the specified key
        # sorted_packages = sorted(
        #     packages,
        #     key=lambda x: x.get(sort_key, 0),
        #     reverse=True,
        # )
        return sorted_packages[:limit]

    except requests.RequestException:
        return []
    except Exception:
        return []


def modify_requirements(package_name, package_version=None, action="install") -> None:
    """Modify the requirements.txt file to install or uninstall a package.

    Args:
        package_name (str): The name of the package to install or uninstall.
        package_version (str, optional): The version of the package to install. Defaults to None.
        action (str): The action to perform, either 'install' or 'uninstall'.

    Raises:
        FileNotFoundError: If the requirements.txt file does not exist when attempting to read.
    """
    lines = get_requirements_packages(as_set=False)

    # Extract the base package name and optional extras
    base_package_name, *extras = package_name.split("[")
    extras_str = "[" + ",".join(extras) if extras else ""
    package_line = next(
        (line for line in lines if base_package_name == line.split("[")[0].split("==")[0]),
        None,
    )

    if action == "install":
        if package_version is not None:
            new_line = f"{base_package_name}{extras_str}=={package_version}"
        else:
            new_line = f"{base_package_name}{extras_str}"

        if package_line:
            # Replace the line with the same base package name
            lines = [line if base_package_name != line.split("[")[0].split("==")[0] else new_line for line in lines]
        else:
            lines.append(new_line)

    elif action == "uninstall":
        # Remove lines with the same base package name
        lines = [line for line in lines if base_package_name != line.split("[")[0].split("==")[0]]

    # Ensure each line ends with a newline character
    lines = [line + "\n" for line in lines]
    with open("requirements.txt", "w") as f:
        f.writelines(lines)


def is_group(line):
    return "[" in line and "]" in line and '"' not in line[line.index("[") : line.index("]")]


def process_dependencies(line, output_lines) -> None:
    lines = []
    if "[" in line and ("]" not in line or "," not in line):
        start = line.index("[")
        lines.append(line[: start + 1])
        line = line[start + 1 :]

    deps = line.split(",")
    i = 0
    while i < len(deps):
        dep = deps[i]
        if "[" in dep and "]" not in dep:
            while "]" not in dep:
                dep += "," + deps[i + 1]
                i += 1
            lines.append(dep + ",")
            i += 1
            continue
        if dep.endswith("]"):
            dep = dep.replace("]", "")
            lines.append(dep)
            lines.append("]")
            i += 1
            continue
        new_dep = dep + "," if dep != deps[-1] else dep
        if new_dep.strip():
            lines.append(new_dep)
        i += 1

    for line in lines:
        output_lines.append(
            "  " + line.strip() if not line.strip().startswith("[") else line.strip(),
        )


def write_pyproject(data, filename="pyproject.toml") -> None:
    """Write the modified pyproject.toml data back to the file."""
    original_data = Path(filename).read_text()
    try:
        with Path(filename).open("w") as f:
            toml_str = tomlkit.dumps(data)
            inside_dependencies = False
            inside_optional_dependencies = False

            input_lines = toml_str.splitlines()
            output_lines = []
            for input_line in input_lines:
                line = input_line.rstrip()
                if is_group(line):
                    inside_dependencies = False
                    inside_optional_dependencies = "optional-dependencies" in line
                    output_lines.append(line)
                    continue

                if "]" in line and inside_dependencies and "[" not in line:
                    inside_dependencies = False

                if inside_optional_dependencies:
                    process_dependencies(line, output_lines)

                if (
                    "dependencies" in line
                    and "optional-dependencies" not in line
                    and "extra-dependencies" not in line
                    and not inside_optional_dependencies
                ):
                    inside_dependencies = True
                    inside_optional_dependencies = False
                    output_lines.append(line[: line.index("[") + 1])
                    line = line[line.index("[") + 1 :]

                if inside_dependencies and not inside_optional_dependencies:
                    inside_optional_dependencies = False
                    process_dependencies(line, output_lines)

                elif not inside_dependencies and not inside_optional_dependencies:
                    output_lines.append(line)

            for line in output_lines:
                line: str = line.strip()
                line = line.removesuffix(",") if line.endswith("],") else line
                chunks = line.split("==")
                if len(chunks) > 2:
                    line = "==".join(chunks[:2])
                    if not line.endswith('",'):
                        line += '",'

                f.write(line + "\n")
    except Exception:
        with Path(filename).open("w") as f:
            f.write(original_data)


def base_name(package_name):
    """Extract the base package name from a package name with optional extras.

    Args:
        package_name (str): The package name with optional extras.

    Returns:
        str: The base package name without extras.
    """
    return package_name.split("[")[0].split("==")[0]


def name_and_version(package_name, upgrade=False):
    if upgrade:
        version = get_latest_version(base_name(package_name))
        return base_name(package_name), version
    if "==" in package_name:
        return package_name.split("==")
    return package_name, None


def modify_dependencies(dependencies, package_version_str, action):
    """Modify the dependencies list for installing or uninstalling a package.

    Args:
        dependencies (list): List of current dependencies.
        package_version_str (str): Package with version string.
        action (str): Action to perform, either 'install' or 'uninstall'.

    Returns:
        list: Modified list of dependencies.
    """
    dependencies = [dep.strip() for dep in dependencies if base_name(dep) != base_name(package_version_str)]
    if action == "install":
        dependencies.append(package_version_str.strip())
    dependencies.sort(key=lambda x: x.lower())  # Sort dependencies alphabetically
    return dependencies


def modify_pyproject_toml(
    package_name,
    package_version="",
    action="install",
    hatch_env=None,
    dependency_group="dependencies",
    pyproject_path="pyproject.toml",
) -> None:
    """Modify the pyproject.toml file to update dependencies based on action.

    Args:
        package_name (str): The name of the package to install or uninstall.
        package_version (str, optional): The version of the package. Defaults to "".
        action (str): The action to perform, either 'install' or 'uninstall'.
        hatch_env (str, optional): The Hatch environment to use. Defaults to "default".
        dependency_group (str, optional): The group of dependencies to modify. Defaults to "dependencies".
    """
    pyproject_path = Path(pyproject_path)
    if not pyproject_path.exists():
        action = input(
            "pyproject.toml not found. Do you want to create it? (y/n): ",
        ).lower()
        if "y" in action.lower():
            create_project(
                input("Enter project name: "),
                input("Enter author name: "),
                input("Enter project description: "),
            )
        elif "n" in action.lower() and "y" in input("Check parent dirs? (y/n): ").lower():
            for _ in range(3):
                if pyproject_path.exists():
                    break
                pyproject_path = Path("../pyproject.toml")

            if not pyproject_path.exists():
                sys.exit(1)

    with open(pyproject_path) as f:
        content = f.read()
        pyproject = tomlkit.parse(content)

    is_optional = dependency_group != "dependencies"
    is_hatch_env = hatch_env and "tool" in pyproject and "hatch" in pyproject["tool"]
    if hatch_env and not is_hatch_env:
        raise ValueError(
            "Hatch environment specified but hatch tool not found in pyproject.toml.",
        )

    # Prepare the package string with version if provided
    package_version_str = f"{package_name}{('==' + package_version) if package_version else ''}"
    base_project = (
        pyproject.get("tool", {}).get("hatch", {}).get("envs", {}).get(hatch_env, {})
        if is_hatch_env
        else pyproject.get("project", {})
    )
    optional_base = pyproject.get("project").get("optional-dependencies", {})

    if is_optional:
        dependencies = optional_base.get(dependency_group, [])
        optional_base[dependency_group] = modify_dependencies(
            dependencies,
            package_version_str,
            action,
        )

        all_group = optional_base.get("all", [])
        optional_base["all"] = modify_dependencies(
            all_group,
            package_version_str,
            action,
        )
        pyproject["project"]["optional-dependencies"] = optional_base
    else:
        dependencies = base_project.get("dependencies", [])
        base_project["dependencies"] = modify_dependencies(
            dependencies,
            package_version_str,
            action,
        )
    if is_hatch_env:
        pyproject["tool"]["hatch"]["envs"][hatch_env] = base_project
    else:
        pyproject["project"] = base_project

    write_pyproject(pyproject, pyproject_path)


def get_pip_freeze():
    """Get the list of installed packages as a set.

    Returns:
        set: Set of installed packages with their versions.
    """
    result = subprocess.run(
        [sys.executable, "-m", "pip", "freeze", "--local"],
        stdout=subprocess.PIPE,
        check=False,
    )
    return set(result.stdout.decode().splitlines())


def is_package_in_requirements(
    package_name,
    requirements_path="requirements.txt",
) -> bool:
    """Check if a package is listed in the requirements.txt file.

    Args:
        package_name (str): The name of the package.

    Returns:
        bool: True if the package is listed in requirements.txt, False otherwise.
    """
    if not Path(requirements_path).exists():
        raise FileNotFoundError("requirements.txt file not found.")
    with Path(requirements_path).open() as f:
        return any(base_name(package_name) == base_name(line) for line in f)


def get_requirements_packages(requirements="requirements.txt", as_set=True):
    """Get the list of packages from the requirements.txt file.

    Returns:
        set: Set of packages listed in the requirements.txt file.
    """
    with Path(requirements).open() as f:
        lines = [line.strip() for line in f if line.strip() and not line.strip().startswith("#")]
    return set(lines) if as_set else lines


def is_package_in_pyproject(
    package_name,
    hatch_env=None,
    pyproject_path="pyproject.toml",
) -> bool:
    """Check if a package is listed in the pyproject.toml file for a given Hatch environment.

    Args:
        package_name (str): The name of the package.
        hatch_env (str): The Hatch environment to check in. Defaults to "default".

    Returns:
        bool: True if the package is listed in pyproject.toml, False otherwise.
    """
    if not Path(pyproject_path).exists():
        raise FileNotFoundError("pyproject.toml file not found.")
    with Path(pyproject_path).open() as f:
        content = f.read()
        pyproject = tomlkit.parse(content)
    is_hatch_env = hatch_env and "tool" in pyproject and "hatch" in pyproject["tool"]
    if hatch_env and not is_hatch_env:
        raise ValueError(
            "Hatch environment specified but hatch tool not found in pyproject.toml.",
        )
    if is_hatch_env:
        dependencies = pyproject["tool"]["hatch"]["envs"][hatch_env]["dependencies"]
    else:
        dependencies = pyproject.get("dependencies", [])
    return any(package_name in dep for dep in dependencies)



"""
Command line interface implementation for pyclean.
"""

import argparse
import logging
import sys

from . import __version__, compat, modern

log = logging.getLogger(__name__)


def parse_arguments():
    """
    Parse and handle CLI arguments.
    """
    debris_default_topics = ['cache', 'coverage', 'package', 'pytest', 'ruff']
    debris_optional_topics = ['jupyter', 'mypy', 'tox']
    debris_choices = ['all', *debris_default_topics, *debris_optional_topics]
    ignore_default_items = [
        '.git',
        '.hg',
        '.svn',
        '.tox',
        '.venv',
        'node_modules',
        'venv',
    ]

    parser = argparse.ArgumentParser(
        description='Remove byte-compiled files for a package or project.',
    )

    if sys.version_info < (3, 8):  # pragma: no-cover-gt-py37
        parser.register('action', 'extend', compat.ExtendAction)

    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument(
        'directory',
        nargs='+',
        help='directory tree to traverse for byte-code',
    )
    parser.add_argument(
        '-i',
        '--ignore',
        metavar='DIRECTORY',
        action='extend',
        nargs='+',
        default=ignore_default_items,
        help='directory that should be ignored (may be specified multiple times;'
        ' default: %s)' % ' '.join(ignore_default_items),
    )
    parser.add_argument(
        '-d',
        '--debris',
        metavar='TOPIC',
        action='extend',
        nargs='*',
        default=argparse.SUPPRESS,
        choices=debris_choices,
        help='remove leftovers from popular Python development tools'
        ' (may be specified multiple times; optional: all %s; default: %s)'
        % (
            ' '.join(debris_optional_topics),
            ' '.join(debris_default_topics),
        ),
    )
    parser.add_argument(
        '-e',
        '--erase',
        metavar='PATTERN',
        action='extend',
        nargs='+',
        default=[],
        help='delete files or folders matching a globbing pattern (may be specified'
        ' multiple times); this will be interactive unless --yes is used.',
    )
    parser.add_argument(
        '-n',
        '--dry-run',
        action='store_true',
        help='show what would be done',
    )

    verbosity = parser.add_mutually_exclusive_group()
    verbosity.add_argument('-q', '--quiet', action='store_true', help='be quiet')
    verbosity.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        help='be more verbose',
    )

    parser.add_argument(
        '-y',
        '--yes',
        action='store_true',
        help='assume yes as answer for interactive questions',
    )

    args = parser.parse_args()
    init_logging(args)

    if args.yes and not args.erase:
        parser.error('Specifying --yes only makes sense with --erase.')

    if 'debris' in args:
        if 'all' in args.debris:
            args.debris = debris_default_topics + debris_optional_topics
        elif not args.debris:
            args.debris = debris_default_topics
        log.debug('Debris topics to scan for: %s', ' '.join(args.debris))
    else:
        args.debris = []

    log.debug('Ignored directories: %s', ' '.join(args.ignore))

    return args


def init_logging(args):
    """
    Set the log level according to the -v/-q command line options.
    """
    log_level = (
        logging.FATAL if args.quiet else logging.DEBUG if args.verbose else logging.INFO
    )
    log_format = '%(message)s'
    logging.basicConfig(level=log_level, format=log_format)


def main():
    """
    Entry point for CLI application.
    """
    args = parse_arguments()

    try:
        modern.pyclean(args)
    except Exception as err:
        raise SystemExit(err)



"""
Modern, cross-platform, pure-Python pyclean implementation.
"""

import logging
from pathlib import Path

BYTECODE_FILES = ['.pyc', '.pyo']
BYTECODE_DIRS = ['__pycache__']
DEBRIS_TOPICS = {
    'cache': [
        '.cache/**/*',
        '.cache/',
    ],
    'coverage': [
        '.coverage',
        'coverage.json',
        'coverage.lcov',
        'coverage.xml',
        'htmlcov/**/*',
        'htmlcov/',
    ],
    'jupyter': [
        '.ipynb_checkpoints/**/*',
        '.ipynb_checkpoints/',
    ],
    'mypy': [
        '.mypy_cache/**/*',
        '.mypy_cache/',
    ],
    'package': [
        'build/bdist.*/**/*',
        'build/bdist.*/',
        'build/lib/**/*',
        'build/lib/',
        'build/',
        'dist/**/*',
        'dist/',
        'sdist/**/*',
        'sdist/',
        '*.egg-info/**/*',
        '*.egg-info/',
    ],
    'pytest': [
        '.pytest_cache/**/*',
        '.pytest_cache/',
        'pytestdebug.log',
    ],
    'ruff': [
        '.ruff_cache/**/*',
        '.ruff_cache/',
    ],
    'tox': [
        '.tox/**/*',
        '.tox/',
    ],
}


class CleanupRunner:  # pylint: disable=too-few-public-methods
    """Module-level configuration and value store."""

    def __init__(self):
        """Cleanup runner with optional dry-run behavior."""
        self.unlink = None
        self.rmdir = None
        self.ignore = None
        self.unlink_count = None
        self.unlink_failed = None
        self.rmdir_count = None
        self.rmdir_failed = None

    def configure(self, args):
        """Set up runner according to command line options."""
        self.unlink = print_filename if args.dry_run else remove_file
        self.rmdir = print_dirname if args.dry_run else remove_directory
        self.ignore = args.ignore
        self.unlink_count = 0
        self.unlink_failed = 0
        self.rmdir_count = 0
        self.rmdir_failed = 0


log = logging.getLogger(__name__)
Runner = CleanupRunner()


def remove_file(fileobj):
    """Attempt to delete a file object for real."""
    log.debug('Deleting file: %s', fileobj)
    try:
        fileobj.unlink()
        Runner.unlink_count += 1
    except OSError as err:
        log.debug('File not deleted. %s', err)
        Runner.unlink_failed += 1


def remove_directory(dirobj):
    """Attempt to remove a directory object for real."""
    log.debug('Removing directory: %s', dirobj)
    try:
        dirobj.rmdir()
        Runner.rmdir_count += 1
    except OSError as err:
        log.debug('Directory not removed. %s', err)
        Runner.rmdir_failed += 1


def print_filename(fileobj):
    """Only display the file name, used with --dry-run."""
    log.debug('Would delete file: %s', fileobj)
    Runner.unlink_count += 1


def print_dirname(dirobj):
    """Only display the directory name, used with --dry-run."""
    log.debug('Would delete directory: %s', dirobj)
    Runner.rmdir_count += 1


def pyclean(args):
    """Cross-platform cleaning of Python bytecode."""
    Runner.configure(args)

    for dir_name in args.directory:
        dir_path = Path(dir_name)

        log.info('Cleaning directory %s', dir_path)
        descend_and_clean(dir_path, BYTECODE_FILES, BYTECODE_DIRS)

        for topic in args.debris:
            remove_debris_for(topic, dir_path)

        remove_freeform_targets(args.erase, args.yes, dir_path)

    log.info(
        'Total %d files, %d directories %s.',
        Runner.unlink_count,
        Runner.rmdir_count,
        'would be removed' if args.dry_run else 'removed',
    )

    if Runner.unlink_failed or Runner.rmdir_failed:
        log.debug(
            '%d files, %d directories %s not be removed.',
            Runner.unlink_failed,
            Runner.rmdir_failed,
            'would' if args.dry_run else 'could',
        )


def descend_and_clean(directory, file_types, dir_names):
    """
    Walk and descend a directory tree, cleaning up files of a certain type
    along the way. Only delete directories if they are empty, in the end.
    """
    for child in sorted(directory.iterdir()):
        if child.is_file():
            if child.suffix in file_types:
                Runner.unlink(child)
        elif child.is_dir():
            if child.name in Runner.ignore:
                log.debug('Skipping %s', child)
            else:
                descend_and_clean(child, file_types, dir_names)

            if child.name in dir_names:
                Runner.rmdir(child)
        else:
            log.debug('Ignoring %s', child)


def remove_debris_for(topic, directory):
    """
    Clean up debris for a specific topic.
    """
    log.debug('Scanning for debris of %s ...', topic.title())

    for path_glob in DEBRIS_TOPICS[topic]:
        delete_filesystem_objects(directory, path_glob)


def remove_freeform_targets(glob_patterns, yes, directory):
    """
    Remove free-form targets using globbing.

    This is **potentially dangerous** since users can delete everything
    anywhere in their file system, including the entire project they're
    working on. For this reason, the implementation imposes the following
    (user experience-related) restrictions:

    - Deleting (directories) is not recursive, directory contents must be
      explicitly specified using globbing (e.g. ``dirname/**/*``).
    - The user is responsible for the deletion order, so that a directory
      is empty when it is attempted to be deleted.
    - A confirmation prompt for the deletion of every single file system
      object is shown (unless the ``--yes`` option is used, in addition).
    """
    for path_glob in glob_patterns:
        log.debug('Erase file system objects matching: %s', path_glob)
        delete_filesystem_objects(directory, path_glob, prompt=not yes)


def delete_filesystem_objects(directory, path_glob, prompt=False):
    """
    Identifies all pathnames matching a specific glob pattern, and attempts
    to delete them in the proper order, optionally asking for confirmation.

    Implementation Note: We sort the file system objects in *reverse order*
    and first delete *all files* before removing directories. This way we
    make sure that the directories that are deepest down in the hierarchy
    are empty (for both files & directories) when we attempt to remove them.
    """
    all_names = sorted(directory.glob(path_glob), reverse=True)
    dirs = (name for name in all_names if name.is_dir() and not name.is_symlink())
    files = (name for name in all_names if not name.is_dir() or name.is_symlink())

    for file_object in files:
        file_type = 'symlink' if file_object.is_symlink() else 'file'
        if prompt and not confirm('Delete %s %s' % (file_type, file_object)):
            Runner.unlink_failed += 1
            continue
        Runner.unlink(file_object)

    for dir_object in dirs:
        if prompt and not confirm('Remove empty directory %s' % dir_object):
            Runner.rmdir_failed += 1
            continue
        Runner.rmdir(dir_object)


def confirm(message):
    """An interactive confirmation prompt."""
    try:
        answer = input('%s? ' % message)
        return answer.strip().lower() in ['y', 'yes']
    except KeyboardInterrupt:
        msg = 'Aborted by user.'
        raise SystemExit(msg)



from __future__ import annotations

import contextlib
import sys

from typing import Any


PY38 = sys.version_info >= (3, 8)


def decode(string: Any, encodings: list[str] | None = None):
    if not isinstance(string, bytes):
        return string

    encodings = encodings or ["utf-8", "latin1", "ascii"]

    for encoding in encodings:
        with contextlib.suppress(UnicodeEncodeError, UnicodeDecodeError):
            return string.decode(encoding)

    return string.decode(encodings[0], errors="ignore")



import contextlib
import io
from pathlib import Path
import time
from typing import Literal

from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.text import Text


class MarkdownStream:
    r"""Stream Markdown text to the terminal with live updates.
    
    Example:
        import time

        # Step 1: Instantiate the MarkdownStream class
        md_stream = MarkdownStream()

        # Step 2: Update the stream with initial text
        md_stream.update("# Welcome to MarkdownStream\nThis is a live stream of Markdown text.")

        # Simulate some delay
        time.sleep(2)

        # Step 3: Update the stream with more text
        md_stream.update("\n## More Content\nHere is some more content being streamed live.")

        # Simulate some more delay
        time.sleep(2)

        # Step 4: Final update to stop the stream
        md_stream.update("\n### End of Stream\nThank you for watching.", final=True)
    """
    live = None
    when = 0
    min_delay = 0.050

    def __init__(self, mdargs=None, style="default", save=False):
        """Initialize the MarkdownStream class.

        Usage:
            >>> from mdstream import MarkdownStream
            >>> md = MarkdownStream(mdargs=None, style="default")
            >>> md.update("# This is a test", final=True)
            >>> md.update("This is a test", font="code", final=False)

        """
        self.printed = ""

        if mdargs:
            self.mdargs = mdargs
        else:
            self.mdargs = {
                "code_theme": "github-light", 
                "inline_code_theme": "github-light",
                "style": style or "reverse",
            }
        self.style = self.mdargs.get("style", "reverse")
        self.live = Live(Text(""), refresh_per_second=1.0 / self.min_delay)
        self.live.start()
        self.buffer = []
        self.save = save

    def __del__(self):  # noqa: D105
        if self.live:
            self.cleanup()

    def cleanup(self):  # noqa: D105
        with contextlib.suppress(Exception):
            try:
                while len(self.buffer) > 0:
                    text, final, font, kwargs = self.buffer.pop(0)
                    if font == "code":
                        text = f"```\n{text}\n```" if len(text.split("\n")) > 1 else f"`{text}`"
                    elif font == "ul":
                        text = "\n".join(f"- {line}" for line in text.split("\n"))
                    elif font == "ol":
                        text = "\n".join(f"{i+1}. {line}" for i, line in enumerate(text.split("\n")))
                    elif font == "blockquote":
                        text = "\n".join(f"> {line}" for line in text.split("\n"))
                    elif font in ["h1", "h2", "h3", "h4", "h5", "h6"]:
                        text = f"{'#' * int(font[1:])} {text}"
                    self.printed += text
                    # Render the entire accumulated text as Markdown
                    string_io = io.StringIO()
                    console = Console(file=string_io, force_terminal=True, style=self.style)
                    markdown = Markdown(self.printed, **self.mdargs)
                    console.print(markdown)
                    output = string_io.getvalue()
                    self.live.update(Text.from_ansi(output))
            except Exception as e:
                pass
            finally:
                self.live.stop()
                self.live = None
 
    
    def update(self, text, 
          final=False,
          font: Literal["text", "code", "li", "ul", "ol", "blockquote",
           "em", "strong", "a", "h1", "h2", "h3", "h4", "h5", "h6"] = "text", **kwargs) -> None:
        self.buffer.append((text.rstrip("\n") + "\n\n", final, font, kwargs))
        now = time.time()
        if not final and now - self.when < self.min_delay:
            return
        if not final:
            self.when = now
            text, final, font, kwargs = self.buffer.pop(0)
            self.mdargs.update(kwargs)
            if font == "code":
                text = f"```\n{text}\n```" if len(text.split("\n")) > 1 else f"`{text}`"
            elif font == "ul":
                text = "\n".join(f"- {line}" for line in text.split("\n"))
            elif font == "ol":
                text = "\n".join(f"{i+1}. {line}" for i, line in enumerate(text.split("\n")))
            elif font == "blockquote":
                text = "\n".join(f"> {line}" for line in text.split("\n"))
            elif font in ["h1", "h2", "h3", "h4", "h5", "h6"]:
                text = f"{'#' * int(font[1:])} {text}"
            self.printed += text


            # Render the entire accumulated text as Markdown
            string_io = io.StringIO()
            console = Console(file=string_io, force_terminal=True, style=self.style)
            markdown = Markdown(self.printed, **self.mdargs)
            console.print(markdown)
            output = string_io.getvalue()

            # Apply the background color to the entire output
            full_text = Text.from_ansi(output)
            if self.live:
                self.live.update(full_text)
        else:
            self.cleanup()
            if self.save:
                with Path("output.md").open("w") as f:
                    f.write(self.printed)

# Example usage of the MarkdownStream class
if __name__ == "__main__":
    # Step 1: Instantiate the MarkdownStream class
    md_stream = MarkdownStream(style="default")

    # Step 2: Update the stream with initial text
    md_stream.update("# Welcome to MarkdownStream\nThis is a live stream of Markdown text.")
    

    # Step 3: Update the stream with more text
    md_stream.update("\n## More Content\nHere is some more content being streamed live.")


    # Step 4: Final update to stop the stream
    md_stream.update("\n### End of Stream\nThank you for watching.", final=True)
    time.sleep(2)